{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE DisambiguateRecordFields #-}
{-# LANGUAGE DuplicateRecordFields #-}

module Laskutin.CSV (CsvInvoice(..), CsvTransaction(..), TransactionMessage(..), parseCsvFile, invoiceRowFromCsv) where

import Data.Bifunctor (first)
import Data.Bifunctor (second)
import Control.Monad ((>=>))
import Data.HashMap.Strict (delete)
import Data.List.NonEmpty (NonEmpty, nonEmpty)
import Data.MIME (Address, address, renderAddress, defaultCharsets, parse)
import Data.Text (Text)
import Data.Text.Encoding (decodeUtf8, encodeUtf8)
import Data.Time (Day)
import Data.Time.Format.ISO8601 (iso8601ParseM)
import Data.Vector (Vector, toList)
import System.Exit (exitFailure)
import System.IO (stderr, hPutStrLn)

import Data.Csv

import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as LBS
import qualified Data.HashMap.Strict as HM
import qualified Data.Text as T
import qualified Data.List.NonEmpty as NE

import Laskutin.Types

data CsvInvoice = CsvInvoice
    { invoiceRecipient :: Address
    , reference :: Reference
    , isPaid :: Bool
    , rows :: [CsvInvoiceRow]
    } deriving Show

data CsvInvoiceRow = CsvInvoiceRow
    { amount :: Maybe Euro
    , name :: Text
    } deriving Show

data CsvTransaction = CsvTransaction
    { date :: Day
    , transactor :: Text
    , referenceOrMessage :: TransactionMessage
    , transactionType :: Text
    , amount :: Euro
    } deriving Show

data TransactionMessage = Message Text
                        | Reference Reference
                        deriving Show

instance FromField Address where
    parseField field = either (const $ fail $ "Invalid email: " <> show field) pure $ parse (address defaultCharsets) field

instance ToField Address where
    toField = renderAddress

instance FromField Bool where
    parseField "x" = pure True
    parseField _ = pure False

instance ToField Bool where
    toField True = "x"
    toField False = ""

instance FromField Day where
    parseField = iso8601ParseM . T.unpack . decodeUtf8

instance FromField TransactionMessage where
    parseField field' =
        let field = decodeUtf8 field'
         in case readReference $ T.unpack field of
              Just reference -> pure $ Reference reference
              Nothing -> pure $ Message field

instance FromNamedRecord CsvInvoice where
    parseNamedRecord row = CsvInvoice
        <$> row .: emailField
        <*> row .: referenceField
        <*> row .: paidField
        <*> namedRecordToRows (foldr delete row [emailField, referenceField, paidField])

instance ToNamedRecord CsvInvoice where
    toNamedRecord CsvInvoice {invoiceRecipient, reference, isPaid, rows} = namedRecord $
        [ emailField .= invoiceRecipient
        , referenceField .= reference
        , paidField .= isPaid
        ] <> (invoiceRowToField <$> rows)

invoiceRowToField :: CsvInvoiceRow -> (BS.ByteString, BS.ByteString)
invoiceRowToField CsvInvoiceRow {amount, name} = encodeUtf8 name .= amount

instance FromNamedRecord CsvTransaction where
    parseNamedRecord row = CsvTransaction
        <$> row .: encodeUtf8 "Päivämäärä"
        <*> row .: encodeUtf8 "Maksaja tai saaja"
        <*> row .: encodeUtf8 "Viite tai viesti"
        <*> row .: encodeUtf8 "Selite"
        <*> row .: encodeUtf8 "Määrä EUR"

namedRecordToRows :: NamedRecord -> Parser [CsvInvoiceRow]
namedRecordToRows row = pure
    $ fmap (uncurry (flip CsvInvoiceRow) . first decodeUtf8)
    $ HM.toList
    $ fmap (readEuro . decodeUtf8) row

emailField, referenceField, paidField :: BS.ByteString
emailField = "sposti"
referenceField = "viite"
paidField = "maksettu"

handleParseErrors :: Either String (a, Vector b) -> IO (a, [b])
handleParseErrors = either (hPutStrLn stderr >=> const exitFailure) (pure . second toList)

parseCsvFile :: FromNamedRecord a => FilePath -> IO (Header, [a])
parseCsvFile = LBS.readFile >=> handleParseErrors . decodeByName

invoiceRowFromCsv :: CsvInvoiceRow -> Maybe InvoiceRow
invoiceRowFromCsv CsvInvoiceRow {amount, name} =
      maybe Nothing (\amount -> Just InvoiceRow {name, amount}) amount
